import pandas as pd
from typing import Union, Optional
import re

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord

from .MASTSearch import MASTSearch
from . import PACKAGEDIR

pd.options.display.max_rows = 10


class KeplerSearch(MASTSearch):
    """
    Search Class that queries mast for observations performed by the Kepler
    Mission, and returns the results in a convenient table with options to download.
    By default both mission products and HLSPs are returned.

    Parameters
    ----------
    target: Optional[Union[str, tuple[float], SkyCoord]] = None
        The target to search for observations of. Can be provided as a name (string),
        coordinates in decimal degrees (tuple), or Astropy `~astropy.coordinates.SkyCoord` Object.
    obs_table:Optional[pd.DataFrame] = None
        Optionally, can provice a Astropy `~astropy.table.Table` Object from
        AstroQuery `astroquery.mast.Observations.query_criteria` which will be used to construct the observations table
    prod_table:Optional[pd.DataFrame] = None
        Optionally, if you provide an obs_table, you may also provide a products table of assosciated products.  These
        two tables will be concatenated to become the primary joint table of data products.
    table:Optional[pd.DataFrame] = None
        Optionally, may provide an stropy `~astropy.table.Table` Object  that is the already merged joint table of obs_table
        and prod_table.
    search_radius:Optional[Union[float,u.Quantity]] = None
        The radius around the target name/location to search for observations.  Can be provided in arcseconds (float) or as an
        AstroPy `~astropy.units.Quantity` Object
    exptime:Optional[Union[str, int, tuple]] = (0,9999)
        Exposure time to filter observation results on.  Can be provided as a mission-specific string,
        an int which forces an exact match to the time in seconds, or a tuple, which provides a range to filter on.
    mission: Optional[Union[str, list[str]]] = ["Kepler", "K2", "TESS"]
        Mission(s) for which to search for data on
    pipeline:  Optional[Union[str, list[str]]] = ["Kepler", "K2", "SPOC"]
        Pipeline(s) which have produced the observed data
    quarter: Optional[Union[int, list[int]]] = None,
        Kepler Observing Quarter(s) for which to search for data.
    month: Optional[int] = None,
        Observation month for Kepler
    """

    _REPR_COLUMNS = [
        "target_name",
        "pipeline",
        "mission",
        "quarter",
        "exptime",
        "distance",
        "year",
        "description",
    ]

    def __init__(
        self,
        target: [Union[str, tuple[float], SkyCoord]],
        obs_table: Optional[pd.DataFrame] = None,
        prod_table: Optional[pd.DataFrame] = None,
        table: Optional[pd.DataFrame] = None,
        search_radius: Optional[Union[float, u.Quantity]] = None,
        exptime: Optional[Union[str, int, tuple]] = (0, 9999),
        pipeline: Optional[Union[str, list[str]]] = None,
        quarter: Optional[Union[int, list[int]]] = None,
        month: Optional[int] = None,
        hlsp: bool = True,
    ):
        if hlsp is False:
            pipeline = ["Kepler"]
            self.mission_search = ["Kepler"]
        else:
            self.mission_search = ["Kepler", "HLSP"]

        super().__init__(
            target=target,
            mission=self.mission_search,
            obs_table=obs_table,
            prod_table=prod_table,
            table=table,
            search_radius=search_radius,
            exptime=exptime,
            pipeline=pipeline,
            sequence=None,
        )
        if table is None:
            self._add_kepler_mission_product()
            self._get_sequence_number()
            self._sort_Kepler()
            # Can't search mast with quarter/month directly, so filter on that after the fact.
            self.table = self.table[self._filter_kepler(quarter, month)]

    @property
    def HLSPs(self):
        """return a MASTSearch object with self.table only containing High Level Science Products"""
        mask = self.table["mission_product"]
        return self._mask(~mask)

    @property
    def mission_products(self):
        """return a MASTSearch object with self.table only containing Mission Products"""
        mask = self.table["mission_product"]
        return self._mask(mask)

    def _check_exact(self, target):
        """Was a Kepler target ID passed?"""
        return re.match(r"^(kplr|kic) ?(\d+)$", target)

    def _target_to_exact_name(self, target):
        "parse Kepler TIC to exact target name"
        return f"kplr{target.group(2).zfill(9)}"

    #
    def _handle_kbonus(self):
        # KBONUS times are masked as they are invalid for the quarter data
        # kbonus_mask = self.table["pipeline"] == "KBONUS-BKG"

        # self.table['start_time'][kbonus_mask] = something
        raise NotImplementedError

    def _add_kepler_mission_product(self):
        """Determine whick products are HLSPs and which are mission products"""
        mission_product = np.zeros(len(self.table), dtype=bool)
        mission_product[self.table["pipeline"] == "Kepler"] = True
        self.table["mission_product"] = mission_product

    def _get_sequence_number(self):
        # Kepler sequence_number values were not populated at the time of
        # writing this code, so we parse them from the description field.
        seq_num = self.table["sequence_number"].values.astype(str)

        mask = (
            (self.table["project"] == "Kepler")
            & self.table["sequence_number"].isna()
            & ~self.table["description"].str.contains("Data Validation")
        )
        re_expr = r".*Q(\d+)"
        seq_num[mask] = [
            re.findall(re_expr, item[0])[0] if re.findall(re_expr, item[0]) else ""
            for item in self.table.loc[mask, ["description"]].values
        ]

        self.table["sequence_number"] = seq_num
        seq_num = [int(x) if x != "<NA>" else 99 for x in seq_num]
        # Create a 'Quarter' column
        self.table["quarter"] = seq_num

        """self.table["mission"] = [
            f"{proj} - Q{seq}"
            if seq !=  '<NA>' else f"{proj}" for proj, seq in zip(self.table["mission"].values.astype(str), seq_num)  
        ]"""

    def _filter_kepler(
        self,
        quarter: Union[int, list[int]] = None,
        month: Union[int, list[int]] = None,
    ) -> pd.DataFrame:
        """Filter Kepler product by month/quarter
        Returns a boolean mask"""
        import os

        products = self.table

        mask = np.ones(len(products), dtype=bool)

        if sum(mask) == 0:
            return products

        # Identify quarter by the description.
        # This is necessary because the `sequence_number` field was not populated
        # for Kepler prime data at the time of writing this function.
        if quarter is not None:
            quarter_mask = np.zeros(len(products), dtype=bool)
            for q in np.atleast_1d(quarter).tolist():
                quarter_mask += products["description"].str.endswith(f"Q{q}")
            mask &= quarter_mask

        # For Kepler short cadence data the month can be specified
        if month is not None:
            month = np.atleast_1d(month).tolist()
            # Get the short cadence date lookup table.
            table = pd.read_csv(
                os.path.join(PACKAGEDIR, "data", "short_cadence_month_lookup.csv")
            )
            # The following line is needed for systems where the default integer type
            # is int32 (e.g. Windows/Appveyor), the column will then be interpreted
            # as string which makes the test fail.
            table["StartTime"] = table["StartTime"].astype(str)
            # Grab the dates of each of the short cadence files.
            # Make sure every entry has the correct month
            is_shortcadence = mask & products["description"].str.contains("Short")

            for idx in np.where(is_shortcadence)[0]:
                quarter = np.atleast_1d(
                    int(
                        products["description"][idx]
                        .split(" - ")[-1]
                        .replace("-", "")[1:]
                    )
                ).tolist()
                date = (
                    products["dataURI"][idx].split("/")[-1].split("-")[1].split("_")[0]
                )
                # Check if the observation date matches the specified Quarter/month from the lookup table
                if (
                    date
                    not in table["StartTime"][
                        table["Month"].isin(month) & table["Quarter"].isin(quarter)
                    ].values
                ):
                    mask[idx] = False
        return mask

    def _sort_Kepler(self):
        sort_priority = {
            "Kepler": 1,
            "KBONUS-BKG": 2,
        }
        df = self.table
        df["sort_order"] = self.table["pipeline"].map(sort_priority).fillna(9)
        df = df.sort_values(
            by=["distance", "sort_order", "quarter", "pipeline", "exptime"],
            ignore_index=True,
        )
        self.table = df

    def filter_table(
        self,
        target_name: Union[str, list[str]] = None,
        pipeline: Union[str, list[str]] = None,
        mission: Union[str, list[str]] = None,
        exptime: Union[int, float, tuple[float]] = None,
        distance: Union[float, tuple[float]] = None,
        year: Union[int, list[int], tuple[int]] = None,
        description: Union[str, list[str]] = None,
        filetype: Union[str, list[str]] = None,
        limit: int = None,
        inplace=False,
        quarter: Optional[int] = None,
        month: Optional[int] = None,
    ):
        """
        Filters the search result table by specified parameters

        Parameters
        ----------
        target_name : str, optional
            Name of targets. A list will look for multiple target names.
        pipeline : str or list[str]], optional
            Data pipeline.  A list will look for multiple pipelines.
        mission : str or list[str]], optional
            Mission. A list will look for muliple missions.
        exptime : int or float, tuple[float]], optional
            Exposure Time. A tuple will look for a range of times.
        distance : float or tuple[float]], optional
            Distance. A float searches for products with a distance less than the value given,
            a tuple will search between the given values.
        year : int or list[int], tuple[int]], optional
            Year. A list will look for multiple years, a tuple will look in the range of years.
        description : str or list[str]], optional
            Description of product. A list will look for descriptions containing any keywords given,
            a tuple will look for descriptions containing all the keywords.
        filetype : str or list[str]], optional
            Type of product. A list will look for multiple filetypes.
        quarter : Optional[int], optional
            Kepler observing quarter, by default None
        month : Optional[int], optional
            Kepler observing month, by default None
        limit : int, optional
            how many rows to return, by default None
        inplace : bool, optional
            whether to modify the KeplerSearch inplace, by default False

        Returns
        -------
        KeplerSearch object with updated table or None if `inplace==True`
        """
        mask = np.ones(len(self.table), dtype=bool)
        mask = self._filter(
            target_name=target_name,
            filetype=filetype,
            exptime=exptime,
            distance=distance,
            year=year,
            description=description,
            pipeline=pipeline,
            mission=mission,
        )

        if (quarter is not None) | (month is not None):
            mask = mask & self._filter_kepler(quarter=quarter, month=month)
        if limit is not None:
            cusu = np.cumsum(mask)
            if max(cusu) > limit:
                mask = mask & (cusu <= limit)
        if inplace:
            self.table = self.table[mask].reset_index()
        else:
            return self._mask(mask)



import pandas as pd
from typing import Union, Optional
import re

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord

from .MASTSearch import MASTSearch

pd.options.display.max_rows = 10


class K2Search(MASTSearch):
    """
    Search Class that queries mast for observations performed by the K2
    Mission, and returns the results in a convenient table with options to download.
    By default both mission products and HLSPs are returned.

    Parameters
    ----------
    target: Optional[Union[str, tuple[float], SkyCoord]] = None
        The target to search for observations of. Can be provided as a name (string),
        coordinates in decimal degrees (tuple), or Astropy `~astropy.coordinates.SkyCoord` Object.
    obs_table:Optional[pd.DataFrame] = None
        Optionally, can provice a Astropy `~astropy.table.Table` Object from
        AstroQuery `~astroquery.mast.Observations.query_criteria` which will be used to construct the observations table
    prod_table:Optional[pd.DataFrame] = None
        Optionally, if you provide an obs_table, you may also provide a products table of assosciated products.  These
        two tables will be concatenated to become the primary joint table of data products.
    table:Optional[pd.DataFrame] = None
        Optionally, may provide an astropy `~astropy.table.Table` Object  that is the already merged joint table of obs_table
        and prod_table.
    search_radius:Optional[Union[float,u.Quantity]] = None
        The radius around the target name/location to search for observations.  Can be provided in arcseconds (float) or as an
        AstroPy `~astropy.units.u.Quantity` Object
    exptime:Optional[Union[str, int, tuple]] = (0,9999)
        Exposure time to filter observation results on.  Can be provided as a mission-specific string,
        an int which forces an exact match to the time in seconds, or a tuple, which provides a range to filter on.
    mission: Optional[Union[str, list[str]]] = ["Kepler", "K2", "TESS"]
        Mission(s) for which to search for data on
    pipeline:  Optional[Union[str, list[str]]] = ["Kepler", "K2", "SPOC"]
        Pipeline(s) which have produced the observed data
    campaign: Optional[Union[int, list[int]]]  = None,
        K2 Observing Campaign(s) for which to search for data.
    """

    _REPR_COLUMNS = [
        "target_name",
        "pipeline",
        "mission",
        "campaign",
        "exptime",
        "distance",
        "year",
        "description",
    ]

    def __init__(
        self,
        target: [Union[str, tuple[float], SkyCoord]],
        obs_table: Optional[pd.DataFrame] = None,
        prod_table: Optional[pd.DataFrame] = None,
        table: Optional[pd.DataFrame] = None,
        search_radius: Optional[Union[float, u.Quantity]] = None,
        exptime: Optional[Union[str, int, tuple]] = (0, 9999),
        pipeline: Optional[Union[str, list[str]]] = None,
        campaign: Optional[Union[int, list[int]]] = None,
        hlsp: bool = True,
    ):
        if hlsp is False:
            pipeline = ["K2"]
            self.mission_search = ["K2"]
        else:
            self.mission_search = ["K2", "HLSP"]
        super().__init__(
            target=target,
            mission=self.mission_search,
            obs_table=obs_table,
            prod_table=prod_table,
            table=table,
            search_radius=search_radius,
            exptime=exptime,
            pipeline=pipeline,
            sequence=campaign,
        )

        if table is None:
            self._add_K2_mission_product()
            self._fix_K2_sequence()
            self._sort_K2()

    @property
    def HLSPs(self):
        """return a MASTSearch object with self.table only containing High Level Science Products"""
        mask = self.table["mission_product"]
        return self._mask(~mask)

    @property
    def mission_products(self):
        """return a MASTSearch object with self.table only containing Mission Products"""
        mask = self.table["mission_product"]
        return self._mask(mask)

    def _check_exact(self, target):
        """Was a K2 target ID passed?"""
        return re.match(r"^(ktwo|epic) ?(\d+)$", target)

    def _target_to_exact_name(self, target):
        "parse K2 TIC to exact target name"
        return f"ktwo{target.group(2).zfill(9)}"

    #
    def _add_K2_mission_product(self):
        """Determine whick products are HLSPs and which are mission products"""
        mission_product = np.zeros(len(self.table), dtype=bool)
        mission_product[self.table["pipeline"] == "K2"] = True
        self.table["mission_product"] = mission_product

    def _fix_K2_sequence(self):
        """K2 campaigns 9, 10, and 11 were split into two sections
        # list these separately in the table with suffixes 'a' and 'b'"""
        seq_num = self.table["sequence_number"].values.astype(str)

        mask = self.table["sequence_number"].isin([9, 10, 11])

        for index, row in self.table[mask].iterrows():
            for half, letter in zip([1, 2], ["a", "b"]):
                if f"c{row['sequence_number']}{half}" in row["productFilename"]:
                    seq_num[index] = f"{int(row['sequence_number']):02d}{letter}"

        self.table["campaign"] = seq_num

    def _sort_K2(self):
        # No specific preference for K2 HLSPs
        sort_priority = {
            "K2": 1,
        }
        df = self.table
        df["sort_order"] = self.table["pipeline"].map(sort_priority).fillna(9)
        df = df.sort_values(
            by=["distance", "sort_order", "campaign", "pipeline", "exptime"],
            ignore_index=True,
        )
        self.table = df

    def filter_table(
        self,
        target_name: Union[str, list[str]] = None,
        pipeline: Union[str, list[str]] = None,
        mission: Union[str, list[str]] = None,
        exptime: Union[int, float, tuple[float]] = None,
        distance: Union[float, tuple[float]] = None,
        year: Union[int, list[int], tuple[int]] = None,
        description: Union[str, list[str]] = None,
        filetype: Union[str, list[str]] = None,
        campaign: Union[int, list] = None,
        limit: int = None,
        inplace=False,
    ):
        """
        Filters the search result table by specified parameters

        Parameters
        ----------
        target_name : str, optional
            Name of targets. A list will look for multiple target names.
        pipeline : str or list[str]], optional
            Data pipeline.  A list will look for multiple pipelines.
        mission : str or list[str]], optional
            Mission. A list will look for muliple missions.
        exptime : int or float, tuple[float]], optional
            Exposure Time. A tuple will look for a range of times.
        distance : float or tuple[float]], optional
            Distance. A float searches for products with a distance less than the value given,
            a tuple will search between the given values.
        year : int or list[int], tuple[int]], optional
            Year. A list will look for multiple years, a tuple will look in the range of years.
        description : str or list[str]], optional
            Description of product. A list will look for descriptions containing any keywords given,
            a tuple will look for descriptions containing all the keywords.
        filetype : str or list[str]], optional
            Type of product. A list will look for multiple filetypes.
        campaign : Optional[int], optional
            K2 observing campaign, by default None
        limit : int, optional
            how many rows to return, by default None
        inplace : bool, optional
            whether to modify the KeplerSearch inplace, by default False

        Returns
        -------
        K2Search object with updated table or None if `inplace==True`
        """
        mask = self._filter(
            target_name=target_name,
            filetype=filetype,
            exptime=exptime,
            distance=distance,
            year=year,
            description=description,
            pipeline=pipeline,
            mission=mission,
            sequence_number=campaign,
        )
        if limit is not None:
            cusu = np.cumsum(mask)
            if max(cusu) > limit:
                mask = mask & (cusu <= limit)
        if inplace:
            self.table = self.table[mask].reset_index()
        else:
            return self._mask(mask)



from astroquery.mast import Observations
import pandas as pd
from typing import Union, Optional
import re
import os

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astroquery.mast import Tesscut

from tqdm import tqdm

from copy import deepcopy

from .MASTSearch import MASTSearch
from . import conf, config, log

PREFER_CLOUD = conf.PREFER_CLOUD
DOWNLOAD_CLOUD = conf.DOWNLOAD_CLOUD

pd.options.display.max_rows = 10


class TESSSearch(MASTSearch):
    """
    Search Class that queries mast for observations performed by the TESS
    Mission, and returns the results in a convenient table with options to download.
    By default mission products and HLSPs are returned.

    Parameters
    ----------
    target: Optional[Union[str, tuple[float], SkyCoord]] = None
        The target to search for observations of. Can be provided as a name (string),
        coordinates in decimal degrees (tuple), or Astropy `~astropy.coordinates.SkyCoord` Object.
    obs_table:Optional[pd.DataFrame] = None
        Optionally, can provice a Astropy `~astropy.table.Table` Object from
        AstroQuery `astroquery.mast.Observations.query_criteria` which will be used to construct the observations table
    prod_table:Optional[pd.DataFrame] = None
        Optionally, if you provide an obs_table, you may also provide a products table of assosciated products.  These
        two tables will be concatenated to become the primary joint table of data products.
    table:Optional[pd.DataFrame] = None
        Optionally, may provide an astropy `~astropy.table.Table` Object  that is the already merged joint table of obs_table
        and prod_table.
    search_radius:Optional[Union[float,u.Quantity]] = None
        The radius around the target name/location to search for observations.  Can be provided in arcseconds (float) or as an
        astropy `~astropy.units.Quantity` Object
    exptime:Optional[Union[str, int, tuple]] = (0,9999)
        Exposure time to filter observation results on.  Can be provided as a mission-specific string,
        an int which forces an exact match to the time in seconds, or a tuple, which provides a range to filter on.
    mission: Optional[Union[str, list[str]]] = ["Kepler", "K2", "TESS"]
        Mission(s) for which to search for data on
    pipeline:  Optional[Union[str, list[str]]] = ["Kepler", "K2", "SPOC"]
        Pipeline(s) which have produced the observed data
    sector: Optional[Union[int, list[int]]] = None,
        TESS Observing Sector(s) for which to search for data.
    """

    _REPR_COLUMNS = [
        "target_name",
        "pipeline",
        "mission",
        "sector",
        "exptime",
        "distance",
        "year",
        "description",
    ]

    def __init__(
        self,
        target: Optional[Union[str, tuple[float], SkyCoord]] = None,
        obs_table: Optional[pd.DataFrame] = None,
        prod_table: Optional[pd.DataFrame] = None,
        table: Optional[pd.DataFrame] = None,
        search_radius: Optional[Union[float, u.Quantity]] = None,
        exptime: Optional[Union[str, int, tuple]] = (0, 9999),
        pipeline: Optional[Union[str, list[str]]] = None,
        sector: Optional[Union[int, list[int]]] = None,
        hlsp: bool = True,
    ):
        if hlsp is False:
            pipeline = ["SPOC", "TESS-SPOC", "TESScut"]
            self.mission_search = ["TESS"]
        else:
            self.mission_search = ["TESS", "HLSP"]

        super().__init__(
            target=target,
            mission=self.mission_search,
            obs_table=obs_table,
            prod_table=prod_table,
            table=table,
            search_radius=search_radius,
            exptime=exptime,
            pipeline=pipeline,
            sequence=sector,
        )

        if table is None:
            if ("TESScut" in np.atleast_1d(pipeline)) or (type(pipeline) is type(None)):
                self._add_tesscut_products(sector)
            self._add_TESS_mission_product()
            self._sort_TESS()

    @property
    def HLSPs(self):
        """return a MASTSearch object with self.table only containing High Level Science Products"""
        mask = self.table["mission_product"]
        return self._mask(~mask)

    @property
    def mission_products(self):
        """return a MASTSearch object with self.table only containing Mission Products"""
        mask = self.table["mission_product"]
        return self._mask(mask)

    @property
    def tesscut(self):
        """return the TESScut only data"""
        mask = self.table["pipeline"] == "TESScut"
        return self._mask(mask)

    @property
    def cubedata(self):
        """return a MASTSearch object with self.table only containing products that are image cubes"""
        mask = self._mask_product_type("target pixel") | (
            self.table["pipeline"] == "TESScut"
        )

        # return self._cubedata()
        return self._mask(mask)

    def _check_exact(self, target):
        """Was a TESS target ID passed?"""
        return re.match(r"^(tess|tic) ?(\d+)$", target)

    def _target_to_exact_name(self, target):
        "parse TESS TIC to exact target name"
        return f"{target.group(2)}"

    def _add_TESS_mission_product(self):
        """Determine whick products are HLSPs and which are mission products"""
        mission_product = np.zeros(len(self.table), dtype=bool)
        mission_product[self.table["pipeline"] == "SPOC"] = True
        self.table["mission_product"] = mission_product
        self.table["sector"] = self.table["sequence_number"]

    def _add_tesscut_products(self, sector_list: Union[int, list[int]]):
        """Add tesscut product information to the search table

        Parameters
        ----------
        sector_list : int, list[int]
            list of sectors to search for tesscut observations of
        """

        # get the ffi info for the targets
        tesscut_info = self._get_tesscut_info(sector_list)
        # add the ffi info to the table
        self.table = pd.concat([self.table, tesscut_info])

    # FFIs only available when using TESSSearch.
    # Use TESS WCS to just return a table of sectors and dates?
    # Then download_ffi requires a sector and time range?

    def _get_tesscut_info(self, sector_list: Union[int, list[int]] = None):
        """Get the tesscut (TESS FFI) obsering information for self.target
        for a particular sector(s)

        Parameters
        ----------
        sector_list : Union[int, list[int]]
            Sector(s) to search for TESS ffi observations

        Returns
        -------
        tesscut_results: pd.DataFrame
            table containing information on sectors in which TESS FFI data is available
        """

        from tesswcs import pointings

        sector_table = Tesscut.get_sectors(coordinates=self.SkyCoord)

        if sector_list is None:
            sector_list = sector_table["sector"].value
        else:
            sector_list = list(
                set(np.atleast_1d(sector_list)) & set(sector_table["sector"].value)
            )

        tesscut_desc = []
        tesscut_mission = []
        tesscut_tmin = []
        tesscut_tmax = []
        tesscut_exptime = []
        tesscut_seqnum = []
        tesscut_year = []

        for sector in sector_list:
            log.debug(f"Target Observable in Sector {sector}")
            tesscut_desc.append(f"TESS FFI Cutout (sector {sector})")
            tesscut_mission.append(f"TESS Sector {sector:02d}")
            tesscut_tmin.append(
                pointings[sector - 1]["Start"] - 2400000.5
            )  # Time(row[5], format="jd").iso)
            tesscut_tmax.append(
                pointings[sector - 1]["End"] - 2400000.5
            )  # Time(row[6], format="jd").iso)
            tesscut_exptime.append(self._sector2ffiexptime(sector))
            tesscut_seqnum.append(sector)
            tesscut_year.append(
                int(
                    np.floor(
                        Time(pointings[sector - 1]["Start"], format="jd").decimalyear
                    )
                )
            )

        # Build the FFI dataframe from the observability
        n_results = len(tesscut_seqnum)
        tesscut_result = pd.DataFrame(
            {
                "description": tesscut_desc,
                "mission": tesscut_mission,
                "target_name": [self.target_search_string] * n_results,
                "targetid": [self.target_search_string] * n_results,
                "t_min": tesscut_tmin,
                "t_max": tesscut_tmax,
                "exptime": tesscut_exptime,
                "productFilename": ["TESScut"] * n_results,
                "provenance_name": ["TESScut"] * n_results,
                "pipeline": ["TESScut"] * n_results,
                "distance": [0] * n_results,
                "sequence_number": tesscut_seqnum,
                "project": ["TESS"] * n_results,
                "obs_collection": ["TESS"] * n_results,
                "year": tesscut_year,
            }
        )

        if len(tesscut_result) > 0:
            log.debug(f"Found {n_results} matching cutouts.")
        else:
            log.debug("Found no matching cutouts.")

        return tesscut_result

    def _sector2ffiexptime(self, sector: Union[int, list[int]]):
        """lookup table for exposure time based off of sector number

        Parameters
        ----------
        sector : Union[int, list[int]]
            sector(s) to get exposure times for

        Returns
        -------
        int
            exposure time in seconds
        """

        # Determine what the FFI cadence was based on sector
        if sector < 27:
            return 1800
        elif (sector >= 27) & (sector <= 55):
            return 600
        elif sector >= 56:
            return 200

    def _sort_TESS(self):
        """Sort Priority for TESS Observations"""
        # Sort TESS results so that SPOC products appear at the top
        sort_priority = {
            "SPOC": 1,
            "TESS-SPOC": 2,
            "TESScut": 3,
        }

        df = self.table
        df["sort_order"] = df["pipeline"].map(sort_priority).fillna(9)
        df = df.sort_values(
            by=["distance", "sort_order", "sector", "pipeline", "exptime"],
            ignore_index=True,
        )
        self.table = df

    def search_sector_ffis(
        self,
        # tmin: Union[float, Time, tuple] = None,
        # tmax: Union[float, Time, tuple] = None,
        # search_radius: Union[float, u.Quantity] = 0.0001 * u.arcsec,
        # exptime: Union[str, int, tuple] = (0, 9999),
        sector: Union[int, type[None]],  # = None,
        **extra_query_criteria,
    ):
        """Returns a list of the FFIs available in a particular sector

        Parameters
        ----------
        sector : Union[int, type[None]]
            sector(s) in which to search for FFI files, by default None

        Returns
        -------
        TESSSearch
            TESSSearch object that contains a joint table of FFI info
        """

        query_criteria = {"project": "TESS", **extra_query_criteria}
        query_criteria["provenance_name"] = "SPOC"
        query_criteria["dataproduct_type"] = "image"

        if sector is not None:
            query_criteria["sequence_number"] = sector

        ffi_obs = Observations.query_criteria(
            objectname=self.target_search_string,
            **query_criteria,
        )

        ffi_products = Observations.get_product_list(ffi_obs)
        # filter out uncalibrated FFIs & theoretical potential HLSP
        prod_mask = ffi_products["calib_level"] == 2
        ffi_products = ffi_products[prod_mask]

        new_table = deepcopy(self)

        # Unlike the other products, ffis don't map cleanly via obs_id as advertised, so we invert and add specific column info
        new_table.obs_table = ffi_products.to_pandas()
        new_table.obs_table["year"] = np.nan

        new_table.prod_table = ffi_obs.to_pandas()
        new_table.table = None

        test_table = new_table._join_tables()
        test_table.reset_index(inplace=True)
        new_table.table = new_table._update_table(test_table)
        new_table.table.reset_index(inplace=True)
        new_table.table["target_name"] = new_table.obs_table["obs_id"]
        new_table.table["obs_collection"] = ["TESS"] * len(new_table.table)
        new_table.table["pipeline"] = [
            new_table.prod_table["provenance_name"].values[0]
        ] * len(new_table.table)
        new_table.table["exptime"] = new_table.table["obs_id"].apply(
            (lambda x: self._sector2ffiexptime(int(x.split("-")[1][1:])))
        )
        new_table.table["year"] = new_table.table["obs_id"].apply(
            (lambda x: int(x.split("-")[0][4:8]))
        )
        new_table.table["sector"] = new_table.table["obs_id"].apply(
            lambda x: int(x.split("-")[1][1:])
        )
        new_table.table["t_min"] = pd.NA * len(new_table.table)
        new_table.table["t_max"] = pd.NA * len(new_table.table)

        return new_table

    def filter_table(
        self,
        target_name: Union[str, list[str]] = None,
        pipeline: Union[str, list[str]] = None,
        mission: Union[str, list[str]] = None,
        exptime: Union[int, float, tuple[float]] = None,
        distance: Union[float, tuple[float]] = None,
        year: Union[int, list[int], tuple[int]] = None,
        description: Union[str, list[str]] = None,
        filetype: Union[str, list[str]] = None,
        limit: int = None,
        inplace=False,
        sector: Union[int, list[str]] = None,
    ):
        """
        Filters the search result table by specified parameters

        Parameters
        ----------
        target_name : str, optional
            Name of targets. A list will look for multiple target names.
        pipeline : str or list[str]], optional
            Data pipeline.  A list will look for multiple pipelines.
        mission : str or list[str]], optional
            Mission. A list will look for muliple missions.
        exptime : int or float, tuple[float]], optional
            Exposure Time. A tuple will look for a range of times.
        distance : float or tuple[float]], optional
            Distance. A float searches for products with a distance less than the value given,
            a tuple will search between the given values.
        year : int or list[int], tuple[int]], optional
            Year. A list will look for multiple years, a tuple will look in the range of years.
        description : str or list[str]], optional
            Description of product. A list will look for descriptions containing any keywords given,
            a tuple will look for descriptions containing all the keywords.
        filetype : str or list[str]], optional
            Type of product. A list will look for multiple filetypes.
        sector : Optional[int], optional
            TESS observing sector, by default None
        limit : int, optional
            how many rows to return, by default None
        inplace : bool, optional
            whether to modify the KeplerSearch inplace, by default False

        Returns
        -------
        TESSSearch object with updated table or None if `inplace==True`
        """
        mask = self._filter(
            target_name=target_name,
            filetype=filetype,
            exptime=exptime,
            distance=distance,
            year=year,
            description=description,
            pipeline=pipeline,
            sequence_number=sector,
            mission=mission,
        )

        if limit is not None:
            cusu = np.cumsum(mask)
            if max(cusu) > limit:
                mask = mask & (cusu <= limit)

        if inplace:
            self.table = self.table[mask].reset_index()
        else:
            return self._mask(mask)

    def download(
        self,
        cloud: bool = conf.PREFER_CLOUD,
        cache: bool = True,
        cloud_only: bool = conf.CLOUD_ONLY,
        download_dir: str = config.get_cache_dir(),
        # TESScut_product="SPOC",
        TESScut_size: Union[int, tuple] = 10,
    ):
        """downloads products in self.table to the local hard-drive

        Parameters
        ----------
        cloud : bool, optional
            enable cloud (as opposed to MAST) downloading, by default True
        cloud_only : bool, optional
            download only products availaible in the cloud, by default False
        download_dir : str, optional
            directory where the products should be downloaded to,
            by default default_download_dir
            cache : bool, optional
        passed to `~astroquery.mast.Observations.download_products`, by default True
            if False, will overwrite the file to be downloaded (for example to replace a corrrupted file)
        remove_incomplete: str, optional
            remove files with a status not "COMPLETE" in the manifest, by default True
        TESScut_size : Union[int, tuple], optional,
            The size of a TESScut FFI cutout in pixels

        Returns
        -------
        ~pandas.DataFrame
            table where each row is an ~astroquery.mast.Observations.download_products()
            manifest

        """

        mast_mf = []
        tesscut_mf = []
        manifest = []
        if "TESScut" not in self.table.provenance_name.unique():
            mast_mf = super().download(
                cloud=cloud,
                cache=cache,
                cloud_only=cloud_only,
                download_dir=download_dir,
            )

        elif "TESScut" in self.table.provenance_name.unique():
            TESSCut_dir = f"{download_dir}/mastDownload/TESSCut"
            if not os.path.isdir(TESSCut_dir):
                os.makedirs(TESSCut_dir)
            mask = self.table["provenance_name"] == "TESScut"
            sector_list = self.table.loc[mask]["sequence_number"].values
            if np.any(~mask):
                mast_mf = self._mask(~mask).download()

            # if cloud:
            #    Tesscut.enable_cloud_dataset()
            tesscut_mf = [
                Tesscut.download_cutouts(
                    coordinates=self.SkyCoord,
                    size=TESScut_size,
                    sector=sector,
                    # Uncomment when astroquery 0.4.8 is released to enable TICA support
                    # product=TESScut_product,
                    # verbose=False
                    path=f"{download_dir}/mastDownload/TESSCut",
                    inflate=True,
                    moving_target=False,  # this could be added
                    mt_type=None,
                ).to_pandas()
                # for sector in sector_list
                for sector in tqdm(
                    sector_list, total=len(sector_list), desc="Downloading TESScut"
                )
            ]
        if len(np.atleast_1d(mast_mf)) != 0:
            manifest = mast_mf

        if len(tesscut_mf) != 0:
            tesscut_mf = pd.concat(tesscut_mf, ignore_index=True)
            # Check to see if files exist, is so mark complete
            tesscut_mf["Status"] = tesscut_mf["Local Path"].apply(
                lambda x: "COMPLETE" if os.path.isfile(x) else "504"
            )

            if len(manifest) != 0:
                manifest = pd.concat([manifest, tesscut_mf], ignore_index=True)
            else:
                manifest = tesscut_mf

        return manifest



from astroquery.mast import Observations
from astroquery.mast import MastClass
import pandas as pd
from typing import Union, Optional
import re
import logging
import warnings
import os

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.time import Time
from tqdm import tqdm

from copy import deepcopy

from .utils import SearchError, SearchWarning, suppress_stdout

from . import conf, config, log

pd.options.display.max_rows = 10


class MASTSearch(object):
    """
    Generic Search Class for data exploration that queries mast for observations performed by the: Kepler, K2, TESS
    Missions, and returns the results in a convenient table with options to download.
    By default only mission products are returned.

    Parameters
    ----------
    target: Optional[Union[str, tuple[float], SkyCoord]] = None
        The target to search for observations of. Can be provided as a name (string),
        coordinates in decimal degrees (tuple), or astropy `~astropy.coordinates.SkyCoord` Object.
    obs_table:Optional[pd.DataFrame] = None
        Optionally, can provice a Astropy `~astropy.table.Table` Object from
        AstroQuery `~astroquery.mast.Observations.query_criteria` which will be used to construct the observations table
    prod_table:Optional[pd.DataFrame] = None
        Optionally, if you provide an obs_table, you may also provide a products table of assosciated products.  These
        two tables will be concatenated to become the primary joint table of data products.
    table:Optional[pd.DataFrame] = None
        Optionally, may provide an stropy `~astropy.table.Table` Object  that is the already merged joint table of obs_table
        and prod_table.
    search_radius:Optional[Union[float,u.Quantity]] = None
        The radius around the target name/location to search for observations.  Can be provided in arcsonds (float) or as an
        AstroPy `~astropy.units.u.Quantity` Object
    exptime:Optional[Union[str, int, tuple]] = (0,9999)
        Exposure time to filter observation results on.  Can be provided as a mission-specific string,
        an int which forces an exact match to the time in seconds, or a tuple, which provides a range to filter on.
    mission: Optional[Union[str, list[str]]] = ["Kepler", "K2", "TESS"]
        Mission(s) for which to search for data on
    pipeline:  Optional[Union[str, list[str]]] = ["Kepler", "K2", "SPOC"]
        Pipeline(s) which have produced the observed data
    sequence: Optional[Union[int, list[int]]] = None,
        Mission Specific Survey value that corresponds to Sector (TESS) AND Campaign (K2). Not valid for Kepler.
        Setting sequence is not recommented for MASTSearch.
    """

    _REPR_COLUMNS = [
        "target_name",
        "pipeline",
        "mission",
        "exptime",
        "distance",
        "year",
        "description",
    ]

    table = None

    def __init__(
        self,
        target: Optional[Union[str, tuple[float], SkyCoord]] = None,
        obs_table: Optional[pd.DataFrame] = None,
        prod_table: Optional[pd.DataFrame] = None,
        table: Optional[pd.DataFrame] = None,
        search_radius: Optional[Union[float, u.Quantity]] = None,
        exptime: Optional[Union[str, int, tuple]] = (0, 9999),
        mission: Optional[Union[str, list[str]]] = ["Kepler", "K2", "TESS"],
        pipeline: Optional[Union[str, list[str]]] = ["Kepler", "K2", "SPOC"],
        sequence: Optional[Union[int, list[int]]] = None,
    ):
        self.search_radius = search_radius
        self.search_exptime = exptime
        self.search_mission = np.atleast_1d(mission).tolist()
        if pipeline is not None:
            pipeline = np.atleast_1d(pipeline).tolist()
        self.search_pipeline = pipeline

        if ("kepler" in (m.lower() for m in mission)) & (sequence is not None):
            log.warning(
                "Sequence not valid when searching for Kepler data. Setting sequence to None"
            )
            sequence = None

        self.search_sequence = sequence

        # Legacy functionality - no longer query kic/tic by integer value only
        if isinstance(target, int):
            raise TypeError(
                "Target must be a target name string, (ra, dec) tuple, "
                "or astropy coordinate object"
            )

        self.target = target

        if isinstance(table, type(None)):
            self._searchtable_from_target(target)
            self.table = self._fix_table_times(self.table)

        # If MAST search tables are provided, another MAST search is not necessary
        else:
            self._searchtable_from_table(table, obs_table, prod_table)

        for col in self.table.columns:
            if not hasattr(self, col):
                setattr(self, col, self.table[col])

    def __len__(self):
        """Returns the number of products in the SearchResult table."""
        return len(self.table)

    def __repr__(self):
        if isinstance(self.table, pd.DataFrame):
            if len(self.table) > 0:
                out = f"{self.__class__.__name__} object containing {len(self.table)} data products \n"
                return out + self.table[self._REPR_COLUMNS].__repr__()
            else:
                return "No results found"
        else:
            return "I am an uninitialized MASTSearch result"

    # Used to call the pandas table html output which is nicer
    def _repr_html_(self):
        if isinstance(self.table, pd.DataFrame):
            if len(self.table) > 0:
                out = f"{self.__class__.__name__} object containing {len(self.table)} data products \n"
                return out + self.table[self._REPR_COLUMNS]._repr_html_()
            else:
                return "No results found"
        else:
            return "I am an uninitialized MASTSearch result"

    def __getitem__(self, key):
        # TODO: Look into class mixins & pandas for this?
        strlist = False
        intlist = False
        if isinstance(key, list):
            if all(isinstance(n, int) for n in key):
                intlist = True
            if all(isinstance(n, str) for n in key):
                strlist = True

        if hasattr(key, "__iter__") or isinstance(key, pd.Series):
            if len(key) == len(self.table):
                return self._mask(key)

        if isinstance(key, (slice, int)) or (intlist):
            if not intlist:
                mask = np.in1d(
                    np.arange(len(self.table)), np.arange(len(self.table))[key]
                )
            else:
                mask = np.in1d(self.table.index, key)
            return self._mask(mask)
        if isinstance(key, str) or strlist:
            # Return a column as a series, or a dataframe of columns
            # Note that we're not returning a Search Object here as
            # we havce additional Requiered columns, etc.
            return self.table[key]

    @property
    def ra(self):
        """Right Ascension coordinate for each data product found."""
        return self.table["s_ra"].values

    @property
    def dec(self):
        """Declination coordinate for each data product found."""
        return self.table["s_dec"].values

    @property
    def exptime(self):
        """Exposure times for all returned products"""
        return self.table["exptime"].values

    @property
    def mission(self):
        """Kepler quarter or TESS sector names for each data product found."""
        return self.table["mission"].values

    @property
    def year(self):
        """Year the observation was made."""
        return self.table["year"].values

    @property
    def pipeline(self):
        """Pipeline name for each data product found."""
        return self.table["pipeline"].values

    @property
    def target_name(self):
        """Target name for each data product found."""
        return self.table["target_name"].values

    @property
    def uris(self):
        """Location Information of the products in the table"""
        uris = self.table["dataURI"].values

        if conf.PREFER_CLOUD:
            cloud_uris = self.cloud_uris
            mask = cloud_uris is not None
            uris[mask] = cloud_uris[mask]

        return uris

    @property
    def cloud_uris(self):
        """Returns the cloud uris for products in self.table.
        Returns
        -------
        ~numpy.array of URI's from ~astroquery.mast
            an array where each element is the cloud-URI of a product in self.table
        """
        if "cloud_uri" not in self.table.columns:
            self.table = self._add_s3_url_column(self.table)

        return self.table["cloud_uri"]

    @property
    def timeseries(self):
        """return a MASTSearch object with self.table only containing products that are a time-series measurement"""
        mask = self._mask_product_type("lightcurve")
        return self._mask(mask)

    @property
    def cubedata(self):
        """return a MASTSearch object with self.table only containing products that are image cubes"""
        mask = self._mask_product_type("target pixel")

        # return self._cubedata()
        return self._mask(mask)

    @property
    def dvreports(self):
        """return a MASTSearch object with self.table only containing products that are data validation pdf files"""
        mask = self._mask_product_type(filetype="dvreport")
        return self._mask(mask)

    # @cached
    def _searchtable_from_target(self, target: Union[str, tuple[float], SkyCoord]):
        """
        takes a target to search and
         - parses that target to search
         - searches mast to create a table from merged Obs/Prod Tables
         - filters the joint table

        Parameters
        ----------
        target : Union[str, tuple[float], SkyCoord]
            the target to search for, either a name (str) or coordinate - (tupe[float]/SkyCoord)

        Returns
        -------
        None - sets self.table equal to the masked/filtered joint table
        """
        self._parse_input(target)

        self.table = self._search(
            search_radius=self.search_radius,
            exptime=self.search_exptime,
            mission=self.search_mission,
            pipeline=self.search_pipeline,
            sequence=self.search_sequence,
        )
        self.table = self._update_table(self.table)

        filetype = [
            "target pixel",
            "lightcurve",
            "dvreport",
        ]
        mask = self._filter(
            exptime=self.search_exptime,
            mission=self.search_mission,
            pipeline=self.search_pipeline,
            sequence_number=self.search_sequence,
            filetype=filetype,
        )
        self.table = self.table[mask].reset_index(drop=True)

    def _searchtable_from_table(
        self,
        table: Optional[pd.DataFrame] = None,
        obs_table: Optional[pd.DataFrame] = None,
        prod_table: Optional[pd.DataFrame] = None,
    ):
        """creates a unified table from either:
            - a passed joint-table
                - this is just passed through
            - an obs_table from astroquery.mast.query_critera.to_pandas()
                - this uses obs_table to create a prod_table and merges tables
            - an obs_table AND a prod_table from from astroquery.mast.get_products.to_pandas()
                - this meges obs_table and prod_table

            self.table is then set from this table
        Parameters
        ----------
        table : Optional[pd.DataFrame], optional
            pre-merged obs_table, prod_table by default None
        obs_table : Optional[pd.DataFrame], optional
            table from astroquery.mast.query_critera.to_pandas(), by default None
        prod_table : Optional[pd.DataFrame], optional
            table from astroquery.mast.get_products.to_pandas(), by default None
        """
        # see if function was passed a joint table
        if isinstance(table, pd.DataFrame):
            self.table = table

        # If we don't have a name or a joint table,
        # check to see if tables were passed
        elif isinstance(obs_table, pd.DataFrame):
            # If we have an obs table and no name, use it
            self.obs_table = obs_table
            if isinstance(prod_table, type(None)):
                # get the prod table if we don't have it
                prod_table = self._search_products(self)
            self.prod_table = prod_table
            self.table = self._join_tables()
        else:
            raise (ValueError("No Target or object table supplied"))

    # def _downsize_table(self, ds_table):
    def _mask(self, mask):
        """Masks down the product and observation tables given an input mask, then returns them as a new Search object.
        deepcopy is used to preserve the class metadata stored in class variables"""

        new_MASTSearch = deepcopy(self)
        new_MASTSearch.table = new_MASTSearch.table[mask].reset_index(drop=True)

        return new_MASTSearch

    def _update_table(self, joint_table: pd.DataFrame):
        """updates self.table for a handfull of convenient column renames and streamlining choices

        Parameters
        ----------
        joint_table : ~pandas.DataFrame
            a dataframe from a merged astroquery.mast.query_criteria & astroquery.mast.get_products tables

        Returns
        -------
        joint_table : ~pandas.DataFrame
            the updated & re-formatted data-frame
        """

        # Modifies a MAST table with user-friendly columns
        if "t_exptime" in joint_table.columns:
            joint_table = joint_table.rename(columns={"t_exptime": "exptime"})
        if "provenance_name" in joint_table.columns:
            joint_table["pipeline"] = joint_table["provenance_name"].copy()
        if "obs_collection_obs" in joint_table.columns:
            joint_table["mission"] = joint_table["obs_collection_obs"].copy()

        # rename identical columns
        joint_table.rename(
            columns={
                "obs_collection_prod": "obs_collection",
                "project_prod": "project",
                "dataproduct_type_prod": "dataproduct_type",
                "proposal_id_prod": "proposal_id",
                "dataRights_prod": "dataRights",
            },
            inplace=True,
        )

        return joint_table

    def _fix_table_times(self, joint_table: pd.DataFrame):
        """fixes incorrect times and adds convenient columns to the search table
        MAST returns the min and max time for each observation in the table.
        Turn this value into the standard JD format
        Also extract the observation year for sorting purposes

        Parameters
        ----------
        joint_table : ~pandas.DataFrame
            the meged obs, prod search table

        Returns
        -------
        ~pandas.DataFrame
            the input table with updated values and additional columns
        """
        if isinstance(joint_table.index, pd.MultiIndex):
            # Multi-Index leading to issues, re-index?
            joint_table = joint_table.reset_index(drop=True)

        year = np.floor(Time(joint_table["t_min"], format="mjd").decimalyear)
        # `t_min` is incorrect for Kepler pipeline products, so we extract year from the filename for those
        for idx, row in joint_table.iterrows():
            if (row["pipeline"] == "Kepler") & (
                "Data Validation" not in row["description"]
            ):
                year[idx] = re.findall(r"\d+.(\d{4})\d+", row["productFilename"])[0]
        joint_table["year"] = year.astype(int)

        joint_table["start_time"] = Time(self.table["t_min"], format="mjd").iso
        joint_table["end_time"] = Time(self.table["t_max"], format="mjd").iso

        return joint_table

    def _search(
        self,
        search_radius: Union[float, u.Quantity] = None,
        exptime: Union[str, int, tuple] = (0, 9999),
        mission: Union[str, list[str]] = ["Kepler", "K2", "TESS"],
        pipeline: Union[str, list[str]] = None,
        sequence: int = None,
    ):
        """from a parsed target input in self.target/self.SkyCoord -
            creates an obs_table & prod_table from self._search_obs() & self._search_prod(,
            and performs an outer joint to merge them

        Parameters
        ----------
        search_radius : Union[float, u.Quantity], optional
            radius (in arcseconds if units not specified)
            around target/coordinates to search, by default None
        exptime : Union[str, int, tuple], optional
            exposure time of products to search for by default (0, 9999)
        pipeline : Union[str, list[str]], optional
            pipeline to search for products from, by default None
        sequence : int, optional
            sequence number (e.g. Cadence/Sector/Campaign) to search for
            products from, by default None

        Returns
        -------
        ~pandas.DataFrame
            A DataFrame resulting from an outer join on a table from
            ~astroquery.mast.Observations.query_criteria and its assosciated
            ~astroquery.mast.Observations.get_product_list
        """
        self.obs_table = self._search_obs(
            search_radius=search_radius,
            exptime=exptime,
            mission=mission,
            pipeline=pipeline,
            sequence=sequence,
            filetype=["lightcurve", "target pixel", "dv"],
        )
        self.prod_table = self._search_prod()
        joint_table = self._join_tables()

        return joint_table

    def _parse_input(self, search_input: Union[str, tuple[float], SkyCoord]):
        """Parses the provided target input search information based on input type
            - If SkyCoord, sets self.Skycoord &
                creates a search string based on coordinates
            - If tuple, assumes RA dec - sets search string based on coordinates &
                creates self.SkyCoord
            - if str, assumes target name - sets search string to input string &
                creates self.SkyCoord from `MastClass().resolve_object()`

        Parameters
        ----------
        search_input : Union[str, tuple, SkyCoord]
            The provided user target search input

        Raises
        ------
        TypeError
            Raise an error if we don't recognise what type of data was passed in
        """
        # If passed a SkyCoord, convert it to an "ra, dec" string for MAST
        self.exact_target = False

        if isinstance(search_input, SkyCoord):
            self.target_search_string = f"{search_input.ra.deg}, {search_input.dec.deg}"
            self.SkyCoord = search_input

        elif isinstance(search_input, tuple):
            self.target_search_string = f"{search_input[0]}, {search_input[1]}"
            self.SkyCoord = SkyCoord(*search_input, frame="icrs", unit="deg")

        elif isinstance(search_input, str):
            self.SkyCoord = MastClass().resolve_object(search_input)
            self.target_search_string = (
                f"{self.SkyCoord.ra.deg}, {self.SkyCoord.dec.deg}"
            )

            target_lower = str(search_input).lower()
            target_str = self._check_exact(target_lower)
            if target_str:
                self.exact_target_name = self._target_to_exact_name(target_str)
                self.exact_target = True

        else:
            raise TypeError(
                "Target must be a target name string or astropy coordinate object"
            )

    def _check_exact(self, target):
        """We dont check exact target name for the generic MAST search -
        TESS/Kepler/K2 have different exact names for identical objects"""
        return False

    def _target_to_exact_name(self, target):
        """We dont check exact target name for the generic MAST search -
        TESS/Kepler/K2 have different exact names for identical objects"""
        return NotImplementedError("Use mission appropriate search for exact targets")

    def _add_s3_url_column(self, joint_table: pd.DataFrame) -> pd.DataFrame:
        """
            self.table will updated to have an extra column of s3 URLS if possible

        Parameters
        ----------
        joint_table : ~pandas.DataFrame
            Dataframe of merged ~astroquery.mast.Observations observations table and product table

        Returns
        -------
        ~pandas.DataFrame
            input dataframe with a column added which countaings the cloud uris of assosciated producs
        """

        logging.getLogger("astroquery").setLevel(log.getEffectiveLevel())

        Observations.enable_cloud_dataset()
        cloud_uris = Observations.get_cloud_uris(
            Table.from_pandas(joint_table.loc[pd.notna(joint_table["dataURI"])]),
            full_url=True,
        )
        joint_table.loc[pd.notna(joint_table["dataURI"]), "cloud_uri"] = cloud_uris
        return joint_table

    def _search_obs(
        self,
        search_radius: Union[float, u.Quantity, None] = None,
        filetype: Union[str, tuple[str]] = ["lightcurve", "target pixel", "dv"],
        mission=["Kepler", "K2", "TESS"],
        pipeline: Union[str, tuple[str], type[None]] = None,
        exptime: Union[int, tuple[int]] = (0, 9999),
        sequence: Union[int, list[int], type[None]] = None,
    ):
        """Assuming we alreads have a pased target search input,
        parse optional inputs and then query mast using
        ~astroquery.mast.Observations.query_criteria

        Parameters
        ----------
        search_radius : Union[float, u.Quantity, None], optional
            radius to search around the target, by default None
        filetype : Union[str, tuple, optional
            type of files to search for, by default ["lightcurve", "target pixel", "dv"]
        mission : list, optional
            mission to search for data from, by default ["Kepler", "K2", "TESS"]
        pipeline : Union[str, tuple, optional
            pipeline to search for data from, by default None
        exptime : Union[int, tuple[int]], optional
            exposure time of data products to search for, by default (0, 9999)
        sequence : Union[int, list[int], type, optional
            mission dependent sequence (e.g. segment/campaign/sector) to search for
            data from, by default None

        Returns
        -------
        ~pandas.DataFrame
            observations table from ~astroquery.mast.Observations.query_criteria().to_pandas()

        Raises
        ------
        SearchError
             - If a ffi filetype is searched, use TESSSearch instead
        SearchError
            If no data is found
        """

        if filetype == "ffi":
            raise SearchError(
                "FFI search not implemented in MASTSearch. Please use TESSSearch."
            )

        # Ensure mission is a list
        mission = np.atleast_1d(mission).tolist()
        if pipeline is not None:
            pipeline = np.atleast_1d(pipeline).tolist()
            # If pipeline "TESS" is used, we assume it is SPOC
            pipeline = np.unique(
                [p if p.lower() != "tess" else "SPOC" for p in pipeline]
            )

        # Speed up by restricting the MAST query if we don't want FFI image data
        # At MAST, non-FFI Kepler pipeline products are known as "cube" products,
        # and non-FFI TESS pipeline products are listed as "timeseries"
        extra_query_criteria = {}
        filetype_query_criteria = {"lightcurve": "timeseries", "target pixel": "cube"}

        extra_query_criteria["dataproduct_type"] = [
            filetype_query_criteria[file.lower()]
            for file in filetype
            if (file.lower() in filetype_query_criteria.keys())
        ]

        observations = self._query_mast(
            search_radius=search_radius,
            project=mission,
            pipeline=pipeline,
            exptime=exptime,
            sequence_number=sequence,
            **extra_query_criteria,
        )
        log.debug(
            f"MAST found {len(observations)} observations. "
            "Now querying MAST for the corresponding data products."
        )
        if len(observations) == 0:
            raise SearchError(f"No data found for target {self.target}.")

        return observations

    def _query_mast(
        self,
        search_radius: Union[float, u.Quantity, None] = None,
        project: Union[str, list[str]] = ["Kepler", "K2", "TESS"],
        pipeline: Union[str, list[str], None] = None,
        exptime: Union[int, float, tuple, type(None)] = (0, 9999),  # None,
        sequence_number: Union[int, list[int], None] = None,
        **extra_query_criteria,
    ):
        """Attempts to find mast observations using ~astroquery.mast.Observations.query_criteria

        Parameters
        ----------
        search_radius : Union[float, u.Quantity, None], optional
            radius around target/location to search, by default None
        project : Union[str, list[str]], optional
            project (mission) to search for data from, by default ["Kepler", "K2", "TESS"]
        pipeline : Union[str, list[str], None], optional
            providence(pipeline) of processed data, by default None
        exptime : Union[int, float, tuple, type, optional
            exposure time of data products to search, by default (0, 9999)
        sequence_number : Union[int, list[int], None], optional
            mission dependent identifier (e.g. Cadence/Campaign/Sector) to search for data products from, by default None

        Returns
        -------
        ~pandas.DataFrame
            an observations_table from query_criteria in pandas DataFrame format

        Raises
        ------
        SearchError
            When unable to resolve search target
        """
        from astroquery.exceptions import NoResultsWarning, ResolverError

        # Constructs the appropriate query for mast
        log.debug(f"Searching for {self.target} with {exptime} on project {project}")

        # We pass the following `query_criteria` to MAST regardless of whether
        # we search by position or target name:
        query_criteria = {"project": project, **extra_query_criteria}
        if pipeline is not None:
            query_criteria["provenance_name"] = pipeline
        if sequence_number is not None:
            query_criteria["sequence_number"] = sequence_number
        if exptime is not None:
            query_criteria["t_exptime"] = exptime

        if self.exact_target and (search_radius is None):
            log.debug(
                f"Started querying MAST for observations with exact name {self.exact_target_name}"
            )

            # do an exact name search with target_name=
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=NoResultsWarning)
                warnings.filterwarnings("ignore", message="t_exptime is continuous")
                obs = Observations.query_criteria(
                    target_name=self.exact_target_name, **query_criteria
                )

            if len(obs) > 0:
                # astroquery does not report distance when querying by `target_name`;
                # we add it here so that the table returned always has this column.
                obs["distance"] = 0.0
                return obs.to_pandas()
        else:
            if search_radius is None:
                search_radius = 0.0001 * u.arcsec

            elif not isinstance(search_radius, u.quantity.Quantity):
                log.warning(
                    f"Search radius {search_radius} units not specified, assuming arcsec"
                )
                search_radius = search_radius * u.arcsec

            query_criteria["radius"] = str(search_radius.to(u.deg))

            try:
                log.debug(
                    "Started querying MAST for observations within "
                    f"{search_radius.to(u.arcsec)} arcsec of objectname='{self.target}'."
                    f"Via {self.target_search_string} search string and query_criteria: "
                    f"{query_criteria}"
                )
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=NoResultsWarning)
                    warnings.filterwarnings("ignore", message="t_exptime is continuous")
                    obs = Observations.query_criteria(
                        objectname=self.target_search_string, **query_criteria
                    )
                obs.sort("distance")
                return obs.to_pandas()
            except ResolverError as exc:
                # MAST failed to resolve the object name to sky coordinates
                raise SearchError(exc) from exc

        return obs.to_pandas()

    def _search_prod(self):
        """uses ~astroquery.mast.Observations.get_product_list to get data products assosicated
        with self.obs_table

        Returns
        -------
        ~pandas.DataFrame
            product table from ~astroquery.mast.Observations.get_product_list.to_pandas
        """
        # Use the search result to get a product list
        products = Observations.get_product_list(Table.from_pandas(self.obs_table))
        return products.to_pandas()

    def _join_tables(self):
        """perform an outer join on self.obs_table on obs_id,
        and self.prod_table and return joined table

        Returns
        -------
        ~pandas.DataFrame
            joined table
        """
        joint_table = pd.merge(
            self.obs_table.reset_index().rename({"index": "obs_index"}, axis="columns"),
            self.prod_table.reset_index().rename(
                {"index": "prod_index"}, axis="columns"
            ),
            on="obs_id",
            how="left",
            suffixes=("_obs", "_prod"),
        ).set_index(["obs_index", "prod_index"])

        log.debug(f"MAST found {len(joint_table)} matching data products.")
        return joint_table

    def _add_kepler_sequence_num(self):
        """adds sequence number to kepler data in the self.table["sequence_number"]
        column since these are not populated in mast
        """
        seq_num = self.table["sequence_number"].values.astype(str)

        # Kepler sequence_number values were not populated at the time of
        # writing this code, so we parse them from the description field.
        mask = (self.table["project"] == "Kepler") & self.table[
            "sequence_number"
        ].isna()
        re_expr = r".*Q(\d+)"
        seq_num[mask] = [
            re.findall(re_expr, item[0])[0] if re.findall(re_expr, item[0]) else ""
            for item in self.table.loc[mask, ["description"]].str.values
        ]

    def _mask_product_type(
        self,
        filetype: str,
    ):
        """convenience function to filter the productFilename column in self.table
        using the pandas .endswith function

        Parameters
        ----------
        filetype : str
            the filetype to filter for

        Returns
        -------
        numpy boolean array
            boolean mask for the column/table
        """

        mask = np.zeros(len(self.table), dtype=bool)
        # This is the dictionary of what files end with that correspond to each allowed file type
        ftype_suffix = {
            "lightcurve": ["lc.fits"],
            "target pixel": ["tp.fits", "targ.fits.gz"],
            "dvreport": ["dvr.pdf", "dvm.pdf", "dvs.pdf"],
        }

        for value in ftype_suffix[filetype]:
            mask |= self.table.productFilename.str.endswith(value)
        return mask

    def _mask_by_exptime(self, exptime: Union[int, tuple[float]]):
        """Helper function to filter by exposure time.
        Returns a boolean array"""
        if "t_exptime" in self.table.columns:
            exposures = self.table["t_exptime"]
        else:
            exposures = self.table["exptime"]

        if isinstance(exptime, (int, float)):
            mask = exposures == exptime
        elif isinstance(exptime, tuple):
            mask = (exposures >= min(exptime)) & (exposures <= max(exptime))
        elif isinstance(exptime, str):
            exptime = exptime.lower()
            if exptime in ["fast"]:
                mask = exposures < 60
            elif exptime in ["short"]:
                mask = (exposures >= 60) & (exposures <= 120)
            elif exptime in ["long", "ffi"]:
                mask = exposures > 120
            elif exptime in ["shortest"]:
                mask = exposures == min(exposures)
            elif exptime in ["longest"]:
                mask = exposures == max(exposures)
            else:
                mask = np.ones(len(exposures), dtype=bool)
                log.debug("invalid string input. No exptime filter applied")
        elif isinstance(exptime, list):
            mask = np.zeros(len(exposures), dtype=bool)
            for et in exptime:
                mask = mask | (exposures == et)

        return mask

    def query_table(
        self,
        criteria: str,
        inplace: bool = False,
    ):
        """Filter the Search Result table using pandas query
        https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.query.html

        Parameters
        ----------
        criteria : str
            string containing criteria to filter table. Can handle multiple criteria,
            e.g. "exptime>=100 and exptime<=500".
        inplace : bool
            if True, modify table in MASTSearch object directly. If False, returns a
            new MASTSearch object with the resulting table

        Returns
        -------
        MASTSearch : MASTSearch object
            Only returned if inplace = False
        """
        filtered_table = self.table.query(criteria).reset_index()
        if inplace:
            self.table = filtered_table
        else:
            new_MASTSearch = deepcopy(self)
            new_MASTSearch.table = filtered_table
            return new_MASTSearch

    def _filter(
        self,
        target_name: Union[str, list[str]] = None,
        pipeline: Union[str, list[str]] = None,
        mission: Union[str, list[str]] = None,
        exptime: Union[int, float, tuple[float]] = None,
        distance: Union[float, tuple[float]] = None,
        year: Union[int, list[int], tuple[int]] = None,
        description: Union[str, list[str], tuple[str]] = None,
        filetype: Union[str, list[str]] = None,
        sequence_number: Union[int, list[int]] = None,
    ):
        """filter self.table based on your product search preferences

        Parameters
        ----------
        target_name : str, optional
            Name of targets.
        pipeline : str or list[str]], optional
            Pipeline provenance to search for data from.
        mission : str or list[str], optional
            Mission projects to search, options are "Kepler", "K2", and "TESS".
        exptime : int or float, tuple[float]], optional
            Exposure time of data products to search.
            An int, float, or list will look for exact matches,
            a tuple will look for a range of times,
            and string values "fast", "short", "long", "shortest", and "longest"
            will match the appropriate exposure times.
        distance : float or tuple[float]], optional
            Distance from given search target to get data products.
            A float searches for products within the given distance,
            a tuple searches between the given values.
        year : int or list[int], tuple[int], optional
            Year of creation.
            A list will look for all years given,
            a tuple will look in the range of years given.
        description : str or list[str], optional
            Key words to search for in the description of the data product.
            A list will look for descriptions containing any keywords given,
            a tuple will look for descriptions containing all the keywords.
        filetype : str or list[str], optional
            file types to search for, options are "target pixel", "lightcurve", "dvreport".
        sequence_number : int or list[int]], optional
            Sequence number to filter by. Corresponds to sector for TESS, campaign for K2, and quarter for Kepler.

        Returns
        -------
        mask : np.ndarray
            cumulative boolean mask for self.table based off of
            the product of individual filter properties
        """
        mask = np.ones(len(self.table), dtype=bool)
        if target_name is not None:
            target_name = np.atleast_1d(target_name).astype(str)
            mask = mask & self.table["target_name"].isin(target_name)

        if not isinstance(filetype, type(None)):
            allowed_ftype = ["lightcurve", "target pixel", "dvreport"]
            filter_ftype = [
                file.lower()
                for file in np.atleast_1d(filetype).astype(str)
                if file.lower() in allowed_ftype
            ]
            if len(filter_ftype) == 0:
                filter_ftype = allowed_ftype
                log.warning("Invalid filetype filtered. Returning all filetypes.")

            file_mask = np.zeros(len(self.table), dtype=bool)
            for ftype in filter_ftype:
                file_mask |= self._mask_product_type(ftype)
            mask = mask & file_mask

        if exptime is not None:
            mask = mask & self._mask_by_exptime(exptime)

        if distance is not None:
            if isinstance(distance, float):
                mask = mask & self.table.eval("distance <= @distance")
            elif isinstance(distance, tuple):
                mask = mask & self.table.eval(
                    "(distance >= @distance[0]) & (distance <= @distance[1])"
                )
            else:
                log.warning(
                    "Invalid input for `distance`, allowed inputs are float and tuple. Ignoring `distance` search parameter."
                )

        if year is not None:
            if isinstance(year, str):
                year = int(year)
            if hasattr(year, "__iter__"):
                year_type = type(year)
                year = [int(y) for y in year]
                year = year_type(year)
            if (
                isinstance(year, np.int_)
                or isinstance(year, int)
                or isinstance(year, list)
            ):
                mask = mask & self.table["year"].isin(np.atleast_1d(year))
            elif isinstance(year, tuple):
                mask = mask & self.table.eval("year>=@year[0] & year<=@year[1]")
            else:
                log.warning(
                    "Invalid input for `year`, allowed inputs are str, int, and tuple. Ignoring `year` search parameter."
                )

        if pipeline is not None:
            pipeline = list(map(str.lower, np.atleast_1d(pipeline).astype(str)))
            mask = mask & self.table["pipeline"].str.lower().isin(pipeline)

        if description is not None:
            if isinstance(description, str):
                mask = mask & self.table["description"].str.lower().str.contains(
                    description.lower()
                )
            elif isinstance(description, tuple):
                # Looks for descriptions which contain *all* of the given keywords
                for word in description:
                    mask = mask & self.table["description"].str.lower().str.contains(
                        word.lower()
                    )
            elif hasattr(description, "__iter__"):
                # Looks for descriptions which contain *any* of the given keywords
                desc_mask = np.zeros(len(self.table), dtype=bool)
                for word in description:
                    desc_mask = desc_mask | self.table[
                        "description"
                    ].str.lower().str.contains(word.lower())
                mask = mask & desc_mask
            else:
                log.warning(
                    "Invalid input for `description`, allowed inputs are str and list[str]. Ignoring `description` search parameter."
                )

        if sequence_number is not None:
            mask = mask & self.table.sequence_number.isin(
                np.atleast_1d(sequence_number)
            )

        if mission is not None:
            mission = list(map(str.lower, np.atleast_1d(mission).astype(str)))
            mask = mask & self.table["mission"].str.lower().isin(mission)

        return mask

    def filter_table(
        self,
        target_name: Union[str, list[str]] = None,
        pipeline: Union[str, list[str]] = None,
        mission: Union[str, list[str]] = None,
        exptime: Union[int, float, tuple[float]] = None,
        distance: Union[float, tuple[float]] = None,
        year: Union[int, list[int], tuple[int]] = None,
        description: Union[str, list[str]] = None,
        filetype: Union[str, list[str]] = None,
        sequence: Union[str, list[str]] = None,
        limit: int = None,
        inplace=False,
    ):
        """Filter the search by keywords

        Parameters
        ----------
        target_name : str, optional
            Name of targets. A list will look for multiple target names.
        pipeline : str or list[str]], optional
            Data pipeline.  A list will look for multiple pipelines.
        mission : str or list[str]], optional
            Mission. A list will look for muliple missions.
        exptime : int or float, tuple[float]], optional
            Exposure Time. A tuple will look for a range of times.
        distance : float or tuple[float]], optional
            Distance. A float searches for products with a distance less than the value given,
            a tuple will search between the given values.
        year : int or list[int], tuple[int]], optional
            Year. A list will look for multiple years, a tuple will look in the range of years.
        description : str or list[str]], optional
            Description of product. A list will look for descriptions containing any keywords given,
            a tuple will look for descriptions containing all the keywords.
        filetype : str or list[str]], optional
            Type of product. A list will look for multiple filetypes.
        sequence : int or list[int]], optional
            Sequence number refers to "quarter" for Kepler, "campaign" for K2, and "sector" for TESS.
        limit : int, optional
            _description_, by default None
        inplace : bool, optional
            _description_, by default False

        Returns
        -------
        MASTSearch or None
            Returns a filtered MASTSearch object or None if `inplace=True`
        """

        mask = self._filter(
            target_name=target_name,
            pipeline=pipeline,
            mission=mission,
            exptime=exptime,
            distance=distance,
            year=year,
            description=description,
            filetype=filetype,
            sequence_number=sequence,
        )

        if limit is not None:
            cusu = np.cumsum(mask)
            if max(cusu) > limit:
                mask = mask & (cusu <= limit)

        if inplace:
            self.table = self.table[mask].reset_index()
        else:
            return self._mask(mask)

    @suppress_stdout
    def _download_one(
        self,
        row: pd.Series,
        cloud_only: bool = False,
        cache: bool = True,
        download_dir: str = ".",
    ) -> pd.DataFrame:
        """Helper function that downloads an individual row.
        This may be more efficient if we are caching, but we can sent a full table
        to download_products to get multiple items.
        """

        # Make sure astroquery uses the same level of verbosity
        logging.getLogger("astropy").setLevel(log.getEffectiveLevel())
        logging.getLogger("astroquery").setLevel(log.getEffectiveLevel())

        # We don't want to query cloud_uri if we don't have to
        # First check to see if we're not downloading on a cloud platform
        # If not - cloud_uris should have already been queried - in that case
        # check to see if a cloud_uri exists, if so we just pass that

        download = True
        if not conf.DOWNLOAD_CLOUD:
            if pd.notna(row["cloud_uri"]):
                download = False
        if conf.DOWNLOAD_CLOUD or download:
            manifest = Observations.download_products(
                Table().from_pandas(row.to_frame(name=" ").transpose()),
                download_dir=download_dir,
                cache=cache,
                cloud_only=cloud_only,
            )
            manifest = manifest.to_pandas()
        else:
            manifest = pd.DataFrame(
                {
                    "Local Path": [row["cloud_uri"]],
                    "Status": ["COMPLETE"],
                    "Message": ["Link to S3 bucket for remote read"],
                    "URL": [None],
                }
            )

        return manifest

    def download(
        self,
        cloud: bool = conf.PREFER_CLOUD,
        cache: bool = True,
        cloud_only: bool = conf.CLOUD_ONLY,
        download_dir: str = config.get_cache_dir(),
        remove_incomplete: str = True,
    ) -> pd.DataFrame:
        """downloads products in self.table to the local hard-drive

        Parameters
        ----------
        cloud : bool, optional
            enable cloud (as opposed to MAST) downloading, by default True
        cloud_only : bool, optional
            download only products availaible in the cloud, by default False
        download_dir : str, optional
            directory where the products should be downloaded to,
            by default default_download_dir
            cache : bool, optional
        passed to `~astroquery.mast.Observations.download_products`, by default True
            if False, will overwrite the file to be downloaded (for example to replace a corrrupted file)
        remove_incomplete: str, optional
            remove files with a status not "COMPLETE" in the manifest, by default True

        Returns
        -------
        ~pandas.DataFrame
            table where each row is an ~astroquery.mast.Observations.download_products()
            manifest

        """

        if len(self.table) == 0:
            warnings.warn("Cannot download from an empty search result.", SearchWarning)
            return None

        if cloud:
            logging.getLogger("astroquery").setLevel(log.getEffectiveLevel())
            Observations.enable_cloud_dataset()

        if (not conf.DOWNLOAD_CLOUD) and ("cloud_uri" not in self.table.columns):
            self.table = self._add_s3_url_column(self.table)

        manifest = [
            self._download_one(row, cloud_only, cache, download_dir)
            for _, row in tqdm(
                self.table.iterrows(),
                total=self.table.shape[0],
                desc="Downloading products",
            )
        ]

        manifest = pd.concat(manifest)
        status = manifest["Status"] != "COMPLETE"
        if np.any(status):
            warnings.warn(
                "Not All Files Downloaded Successfully, Check Returned Manifest.",
                SearchWarning,
            )
            if remove_incomplete:
                for file in manifest.loc[status]["Local Path"].values:
                    if os.path.isfile(file):
                        os.remove(file)
                        warnings.warn(f"Removed {file}", SearchWarning)
                    else:
                        warnings.warn(f"Not a file: {file}", SearchWarning)
        manifest = manifest.reset_index(drop=True)
        return manifest



import os
import warnings
import glob
import shutil
import logging

import astropy.config as _config

ROOTNAME = "lksearch"
log = logging.getLogger(ROOTNAME)


class ConfigNamespace(_config.ConfigNamespace):
    rootname = ROOTNAME


class ConfigItem(_config.ConfigItem):
    rootname = ROOTNAME


def get_config_dir():
    """
    Determines the package configuration directory name and creates the
    directory if it doesn't exist.

    This directory is typically ``$HOME/.lksearch/config``, but if the
    XDG_CONFIG_HOME environment variable is set and the
    ``$XDG_CONFIG_HOME/lksearch`` directory exists, it will be that directory.
    If neither exists, the former will be created and symlinked to the latter.

    Returns
    -------
    configdir : str
        The absolute path to the configuration directory.

    """
    return _config.get_config_dir(ROOTNAME)


def get_config_file():
    return f"{get_config_dir()}/{ROOTNAME}.cfg"


def create_config_file(overwrite: bool = False):
    """Creates a default configuration file in the config directory"""

    from .. import conf

    # check if config file exists
    path_to_config_file = get_config_file()
    cfg_exists = os.path.isfile(path_to_config_file)

    if not cfg_exists or (cfg_exists and overwrite):
        with open(path_to_config_file, "w", encoding="utf-8") as f:
            for item in conf.items():
                f.write(f"## {item[1].description} \n")
                f.write(f"# {item[0]} = {item[1].defaultvalue} \n")
                f.write("\n")
    else:
        log.error("Config file exists and overwrite set to {overwrite}")


def get_cache_dir():
    """
    Determines the default lksearch cache directory name and creates the
    directory if it doesn't exist. If the directory cannot be access or created,
    then it returns the current directory (``"."``).

    This directory is typically ``$HOME/.lksearch/cache``, but if the
    XDG_CACHE_HOME environment variable is set and the
    ``$XDG_CACHE_HOME/lksearch`` directory exists, it will be that directory.
    If neither exists, the former will be created and symlinked to the latter.

    The value can be also configured via ``cache_dir`` configuration parameter.

    Returns
    -------
    cachedir : str
        The absolute path to the cache directory.

    See `~lksearch.Conf` for more information.
    """
    from .. import conf

    cache_dir = conf.cache_dir
    if cache_dir is None or cache_dir == "":
        cache_dir = _config.get_cache_dir(ROOTNAME)
    cache_dir = _ensure_cache_dir_exists(cache_dir)
    cache_dir = os.path.abspath(cache_dir)

    return cache_dir


def _ensure_cache_dir_exists(cache_dir):
    if os.path.isdir(cache_dir):
        return cache_dir
    else:
        # if it doesn't exist, make a new cache directory
        try:
            os.mkdir(cache_dir)
        # user current dir if OS error occurs
        except OSError:
            warnings.warn(
                "Warning: unable to create {} as cache dir "
                " (for downloading MAST files, etc.). Use the current "
                "working directory instead.".format(cache_dir)
            )
            cache_dir = "."
        return cache_dir


def clearcache(test=True):
    """Deletes all downloaded files in the lksearch download directory

    Parameters
    ----------
    test : bool, optional
       perform this in test mode, printing what folders will be deleted, by default True.
       Set test=False to delete cache
    """
    # Check to see if default download dir/mastDownload exists
    mastdir = f"{get_cache_dir()}/mastDownload"
    if os.path.isdir(mastdir):
        files = glob.glob(f"{mastdir}/*")
        if test:
            print("Running in test mode, rerun with test=False to clear cache")
        for f in files:
            if test:
                print(f"removing {f}")
            else:
                shutil.rmtree(f)



import os
import sys
from functools import wraps

from . import config

default_download_dir = config.get_cache_dir()


class SearchError(Exception):
    pass


class SearchWarning(Warning):
    pass


def suppress_stdout(f, *args, **kwargs):
    """A simple decorator to suppress function print outputs."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        # redirect output to `null`
        with open(os.devnull, "w") as devnull:
            old_out = sys.stdout
            sys.stdout = devnull
            try:
                return f(*args, **kwargs)
            # restore to default
            finally:
                sys.stdout = old_out

    return wrapper



#!/usr/bin/env python
from __future__ import absolute_import
from . import config as _config
import logging
import os

__version__ = "1.1.0"
PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))


class Conf(_config.ConfigNamespace):
    """
    Configuration parameters for `search`.

    Refer to `astropy.config.ConfigNamespace` for API details.

    Refer to `Astropy documentation <https://docs.astropy.org/en/stable/config/index.html#accessing-values>`_
    for usage.

    The attributes listed below are the available configuration parameters.

    Parameters
    ----------
    search_result_display_extra_columns
        List of extra columns to be included when displaying a SearchResult object.

    cache_dir
        Default cache directory for data files downloaded, etc. Defaults to ``~/.lksearch/cache`` if not specified.

    PREFER_CLOUD
        Use Cloud-based data product retrieval where available (primarily Amazon S3 buckets for MAST holdings)

    DOWNLOAD_CLOUD
       Download cloud based data. If False, download() will return a pointer to the cloud based data instead of
       downloading it - intended usage for cloud-based science platforms (e.g. TIKE)
    """

    # Note: when using list or string_list datatype,
    # the behavior of astropy's parsing of the config file value:
    # - it does not accept python list literal
    # - it accepts a comma-separated list of string
    #   - for a single value, it needs to be ended with a comma
    # see: https://configobj.readthedocs.io/en/latest/configobj.html#the-config-file-format
    search_result_display_extra_columns = _config.ConfigItem(
        [],
        "List of extra columns to be included when displaying a SearchResult object.",
        cfgtype="string_list",
        module="lksearch",
    )

    cache_dir = _config.ConfigItem(
        None,
        "Default cache directory for data files downloaded, etc.",
        cfgtype="string",
        module="lksearch.config",
    )

    CLOUD_ONLY = _config.ConfigItem(
        False,
        "Only Download cloud based data."
        "If False, will download all data"
        "If True, will only download data located on a cloud (Amazon S3) bucket",
        cfgtype="boolean",
    )

    PREFER_CLOUD = _config.ConfigItem(
        True,
        "Prefer Cloud-based data product retrieval where available",
        cfgtype="boolean",
    )

    DOWNLOAD_CLOUD = _config.ConfigItem(
        True,
        "Download cloud based data."
        "If False, download() will return a pointer to the cloud based data"
        "instead of downloading it - intended usage for cloud-based science platforms (e.g. TIKE)",
        cfgtype="boolean",
    )


conf = Conf()
log = logging.getLogger("lksearch")

from .MASTSearch import MASTSearch  # noqa
from .TESSSearch import TESSSearch  # noqa
from .KeplerSearch import KeplerSearch  # noqa
from .K2Search import K2Search  # noqa

# from .catalogsearch import *  # noqa
from . import catalogsearch  # noqa



"""Catalog module to search various catalogs for missions"""

from typing import Union
import numpy as np

from astropy.coordinates import Angle, SkyCoord, Distance
from astropy.table import Table
from astropy.time import Time
import astropy.units as u

from astroquery.vizier import Vizier
from astroquery.utils.tap.core import TapPlus
from astroquery.mast import MastClass
from astroquery.simbad import Simbad

import pandas as pd

import warnings
import time
import json

from . import log
from . import PACKAGEDIR


# This is a lits of VizieR catalogs and their input parameters to be used in the
# query_skycatalog function
def _load_cat_config():
    with open(f"{PACKAGEDIR}/data/catalog_config.json", "r") as j:
        cat_dict = json.loads(j.read())
    for key in cat_dict.keys():
        cat_dict[key]["equinox"] = Time(
            cat_dict[key]["equinox"], format="jyear", scale="tt"
        )
    return cat_dict


_Catalog_Dictionary = _load_cat_config()

# Connect to the Vizier TAP server here so that we only do this once
VizTap = TapPlus(url="http://TAPVizieR.u-strasbg.fr/TAPVizieR/tap/")

# TODO Swap this out with a configuration parameter maybe?
# Or None and Raise Exception
# Make this an optional keword argument for debugging/doc
_default_catalog = "tic"


def query_region(
    search_input: Union[str, SkyCoord, tuple, list[str, SkyCoord, tuple]],
    output_epoch: Union[str, Time] = None,
    catalog: str = "tic",
    radius: Union[float, u.Quantity] = u.Quantity(100, "arcsecond"),
    magnitude_limit: float = 18.0,
    max_results: int = None,
    return_skycoord: bool = False,
):
    """
    Query a catalog for a single source location, obtain nearby sources

    Parameters
    ----------
    coord : `~astropy.coordinates.SkyCoord`, string, tuple, or list thereof
        Coordinates around which to do a radius query. If passed a string, will first try to resolve string as a coordinate using `~astropy.coordinates.SkyCoord`, if this fails then tries to resolve the string as a name using `~astroquery.mast.MastClass.resolve_object`.
    output_epoch: `~astropy.time.Time`
        The time of observation in JD.
    catalog: str
        The catalog to query, either 'kepler', 'k2', or 'tess', 'gaia'
    radius : float or `~astropy.units.quantity.Quantity`
        Radius in arcseconds to query
    magnitude_limit : float
        A value to limit the results in based on the Tmag/Kepler mag/K2 mag or Gaia G mag. Default, 18.
    return_skycoord: bool
        Whether to return an `~astropy.coordinates.SkyCoord` object. Default is False.

    Returns
    -------
    result: `~astropy.table.Table` or `~astropy.coordinates.SkyCoord`
        By default returns a  pandas dataframe of the sources within radius query, corrected for proper motion. Optionally will return astropy.coordinates.SkyCoord object.

    """

    coord, search_catalog = _parse_search_input(search_input, catalog=catalog)

    # Check to make sure that user input is in the correct format
    if not isinstance(coord, SkyCoord):
        if isinstance(coord, str):
            coord = MastClass().resolve_object(coord)
        else:
            raise TypeError(f"could not resolve {coord} to SkyCoord")
    if output_epoch is not None:
        if not isinstance(output_epoch, Time):
            try:
                output_epoch = Time(output_epoch, format="jd")
            except ValueError:
                raise TypeError(
                    "Must pass an `astropy.time.Time object` or parsable object."
                )
            raise TypeError(
                "Must pass an `astropy.time.Time object` or parsable object."
            )
    if not coord.isscalar:
        raise ValueError("must pass one target only.")

    # Here we check to make sure that the radius entered is in arcseconds
    # This also means we do not need to specify arcseconds in our catalog query
    try:
        radius = u.Quantity(radius, "arcsecond")
    except u.UnitConversionError:
        raise

    # Check to make sure that the catalog provided by the user is valid for this function
    if search_catalog.lower() not in _Catalog_Dictionary.keys():
        raise ValueError(f"Can not parse catalog name '{catalog}'")
    catalog_meta = _Catalog_Dictionary[search_catalog.lower()]

    # Get the Vizier catalog name
    catalog_name = catalog_meta["catalog"]

    # Get the appropriate column names and filters to be applied
    filters = Vizier(
        columns=catalog_meta["columns"],
        column_filters={catalog_meta["column_filters"]: f"<{magnitude_limit}"},
    )
    # The catalog can cut off at 50 - we dont want this to happen
    if max_results is not None:
        filters.ROW_LIMIT = max_results
    else:
        filters.ROW_LIMIT = -1
    # Now query the catalog
    result = filters.query_region(coord, catalog=catalog_name, radius=Angle(radius))
    if len(result) == 0:
        # Make an empty Table
        empty_table = pd.DataFrame(
            columns=[
                *catalog_meta["columns"],
                "ID",
                "RA",
                "Dec",
                "Separation",
                "Relative_Flux",
            ]
        )
        # Make Sure Columns are consistently renamed for the catalog
        empty_table = empty_table.rename(
            {
                i: o
                for i, o in zip(catalog_meta["rename_in"], catalog_meta["rename_out"])
            },
            axis=1,
        )
        # Make sure we have an index set
        empty_table = empty_table.set_index("ID")
        return empty_table

    result = result[catalog_name]
    # Rename the columns so that the output is uniform
    result.rename_columns(
        catalog_meta["rename_in"],
        catalog_meta["rename_out"],
    )
    if catalog_meta["prefix"] is not None:
        prefix = catalog_meta["prefix"]
        result["ID"] = [f"{prefix} {id}" for id in result["ID"]]
    if output_epoch is None:
        output_epoch = catalog_meta["equinox"]
    c = _table_to_skycoord(
        table=result,
        equinox=catalog_meta["equinox"],
        output_epoch=output_epoch,
        catalog=search_catalog,
    )
    ref_index = np.argmin(coord.separation(c).arcsecond)
    sep = c[ref_index].separation(c)
    if return_skycoord:
        s = np.argsort(sep.deg)
        return c[s]
    result["RA"] = c.ra.deg
    result["Dec"] = c.dec.deg
    result["Separation"] = sep.arcsecond
    # Calculate the relative flux
    result["Relative_Flux"] = 10 ** (
        (
            result[catalog_meta["default_mag"]]
            - result[catalog_meta["default_mag"]][ref_index]
        )
        / -2.5
    )
    # Now sort the table based on separation
    result.sort(["Separation"])
    # return result
    # result = result.to_pandas().set_index("ID")
    return result[_get_return_columns(result.columns)].to_pandas()


def query_id(
    search_object: Union[str, int, list[str, int]],
    output_catalog: str = None,
    input_catalog: str = None,
    max_results: int = None,
    return_skycoord: bool = False,
    output_epoch: Union[str, Time] = None,
):
    """Searches a catalog (TIC, KIC, EPIC, or GAIA DR3) for an exact ID match and
    returns the assosciated catalog rows.  A limited cross-match between the TIC, KIC, and gaiadr3
    catalogs is possible using the catalog, and input_catalog optional parameters.

    Parameters
    ----------
    search_object : Union[str, int, list[str, int]]
        A string or integer, or list of strings or integers, that represents
        a list of IDs from a single catalog to match.  If an integer is supplied the
        catalog optional parameter must be specified.
    catalog : str, optional
        Catalog to search for an ID match to.  If no input_catalog is
        specified catalog and input_catalog are assumed to be the same.
        If search_object is a string and catalog and is None, search_object is
        parsed to try and determine the catalog, by default None
    input_catalog : str, optional
        _description_, by default None
    max_results : int, optional
        limits the maximum rows to return, by default None
    return_skycoord : bool, optional
        If true, an `~astropy.coordinates.SkyCoord` objects is returned for each
        row in the result table, by default False
    output_epoch : Union[str, Time], optional
        If a return_skycoord is True, output_epoch can be used to specify the output_epoch for the
        returned SkyCoord object, by default None

    Returns
    -------
    results_table: Union[Table, SkyCoord, list[SkyCoord]]
        `~astropy.table.Table` object containing the rows of the catalog with IDs matching the search_input.
        If return_skycoord is set to True, a `~astropy.coordinates.SkyCoord` object or list of `~astropy.coordinates.SkyCoord` objects
        is instead returned.

    """
    id_column = None

    if isinstance(search_object, list):
        id_list = _parse_id_list(search_object)
    else:
        id_list, scat = _parse_id(search_object)
        # IF we can figure out the soruce catalog from context -
        # EG TIC Blah, assume the catalog to search is the catalog detected
        # And th
        if output_catalog is None and scat is not None:
            output_catalog = scat
        if input_catalog is None and scat is not None:
            input_catalog = scat

    # Assuming a 1:1 match.  TODO is this a bad assumption?
    if max_results is None:
        max_results = len(np.atleast_1d(search_object))

    if output_catalog is not None and input_catalog is not None:
        if output_catalog != input_catalog:
            max_results = max_results * 10
            if input_catalog in np.atleast_1d(
                _Catalog_Dictionary[output_catalog]["crossmatch_catalogs"]
            ):
                if _Catalog_Dictionary[output_catalog]["crossmatch_type"] == "tic":
                    # TIC is is crossmatched with gaiadr3/kic
                    # If KIC data for a gaia source or vice versa is desired
                    # search TIC to get KIC/gaia ids then Search KIC /GAIA
                    source_id_column = _Catalog_Dictionary["tic"][
                        "crossmatch_column_id"
                    ][input_catalog]
                    new_id_table = _query_id(
                        "tic", id_list, max_results, id_column=source_id_column
                    )
                    id_list = ", ".join(
                        new_id_table[
                            _Catalog_Dictionary["tic"]["crossmatch_column_id"][
                                output_catalog
                            ]
                        ].astype(str)
                        # .values
                    )
                if _Catalog_Dictionary[output_catalog]["crossmatch_type"] == "column":
                    # TIC is is crossmatched with gaiadr3/kic
                    # If we want TIC Info for a gaiadr3/KIC source - match appropriate column in TIC
                    id_column = _Catalog_Dictionary[output_catalog][
                        "crossmatch_column_id"
                    ][input_catalog]
            else:
                raise ValueError(
                    f"{input_catalog} does not have crossmatched IDs with {output_catalog}. {output_catalog} can be crossmatched with {_Catalog_Dictionary[catalog]['crossmatch_catalogs']}"
                )
    else:
        if output_catalog is None:
            output_catalog = _default_catalog

    results_table = _query_id(output_catalog, id_list, max_results, id_column=id_column)
    if return_skycoord:
        return _table_to_skycoord(
            results_table, output_epoch=output_epoch, catalog=output_catalog
        )
    else:
        return results_table.to_pandas()


def _query_id(catalog: str, id_list: str, max_results: int, id_column: str = None):
    query = _get_TAP_Query(
        catalog, id_list, max_results=max_results, id_column=id_column
    )
    async_limit = 1e3
    if max_results > async_limit:
        # we should chex max_results and if low do a synchronous query, if large async
        log.warn(
            f"Warning: Queries over {async_limit} will be done asynchronously, and may take some time"
        )
        job = VizTap.launch_job_async(query)
        job.wait_for_job_end()
        results_table = job.get_data()
    else:
        job = VizTap.launch_job(query)
        results_table = job.get_data()
    return results_table  # .to_pandas()


# use simbad to get name/ID crossmatches
def query_names(search_input: Union[str, list[str]]):
    """Uses the Simbad name resolver and ids to disambiguate the search_input string or list.

    Parameters
    ----------
    search_input: Union[str, list[str]]
        A string or list of strings to query simbad for ID disambiguation

    Returns
    -------
    result: DataFrame, list[DataFrame]
        Results from the `~astroquery.simbad.Simbad` ID query in `~pandas.DataFrame` format.

    """
    # match_str = None only usable in bleeding edge astroquery
    # match_list = _Catalog_Dictionary.keys()

    # if match_catalog is not None:
    #    if match_catalog.lower() in match_list:
    #        match_str = _Catalog_Dictionary[match_catalog.lower()]["SIMBAD_match_like"]

    if isinstance(search_input, list):
        result = []
        log.warning("Throttling query limit to Simbad's: max 5/s")
        for item in search_input:
            result_iter = _query_names(item)
            time.sleep(0.2)
            result.append(result_iter)
    else:
        result = _query_names(search_input)

    return result


def match_names_catalogs(
    search_input: Union[str, list[str]], match: Union[str, list[str]]
):
    """Uses the Simbad name resolver and ids to disambiguate the search_input string or list, and compare the disambiguated names with a list to match against.

    Parameters
    ----------
    search_input: Union[str, list[str]]
        A string or list of strings to query simbad for ID disambiguation

    match: Union[str, list[str]]
        Short name of catalog to parse the simbad id results for.  If this is passed the list of ids are not
        reported and a column per item in the match list is added and the ids with that match str contained in the id are listed.

    Returns
    -------
    result: DataFrame, list[DataFrame]
        Results from the `~astroquery.simbad.Simbad` ID query in `~pandas.DataFrame` format.

    """
    result = query_names(search_input)

    # Make an Empty DataFrame with # columns = 1(the search input) + len(match)
    # This will provide the ids that match each search input given the match criteria
    col = ["search"]
    for item in np.atleast_1d(match):
        col.append(item)
    final_result = pd.DataFrame(columns=col, dtype=str)

    # Iterate through the search inputs, returned ids, and match criteria
    i = 0
    for item in np.atleast_1d(search_input):
        row = [item]
        # make sure we're in a list in the event we have a single result
        if not isinstance(result, list):
            result = [result]
        # For each item in the match terms, see if it is contained in ID
        for cat in np.atleast_1d(match):
            mcat = cat.strip().replace(" ", "").lower()
            cmatch = None
            for sid in result[i]["id"]:
                id = sid.strip().replace(" ", "").lower()
                if mcat in id:
                    if cmatch is None:
                        cmatch = sid
                    else:
                        cmatch.append(sid)
            if cmatch is None:
                cmatch = ""
            row.append(cmatch)

        row_result = pd.DataFrame(np.array([row]), columns=col, dtype=str)
        final_result = pd.concat(
            [final_result, row_result],
            ignore_index=True,
            axis=0,
        )
        i += 1
    # get rid of our old list of tables result and return the matching results
    result = final_result

    return result


def _query_names(search_item):
    # Construct exact ID TAP queries for various surveys
    #    result_table = Simbad.query_objectids(search_item, criteria = match_str)
    result_table = Simbad.query_objectids(search_item)
    result_table = result_table.to_pandas()

    # Older astroquery versions return bytes-like objects, lets make sure things are strings
    for col, dtype in result_table.dtypes.items():
        if dtype == object and col == "ID":  # Only process byte object columns
            result_table[col] = result_table[col].str.decode("utf-8")

    # older astroquery versions use "ID" not 'id', lets be backwards compatible
    if "ID" in result_table.keys():
        result_table.rename(columns={"ID": "id"}, inplace=True)

    return result_table


def _get_TAP_Query(catalog: str, ID: str, max_results: int = None, id_column=None):
    if catalog not in _Catalog_Dictionary.keys():
        raise ValueError(f"{catalog} not found in TAP catalogs list")
    if max_results is None:
        max_results = len(ID.split(","))
    if id_column is None:
        id_column = _Catalog_Dictionary[catalog]["default_id_column"]

    query_model = f""" SELECT TOP {max_results} * 
    FROM "{_Catalog_Dictionary[catalog]["catalog"]}" 
    WHERE {id_column} IN({ID}) """

    return query_model


def _parse_search_input(search_input, catalog: str = None):
    if isinstance(search_input, SkyCoord):
        search_object = search_input
    elif isinstance(search_input, tuple):
        search_object = SkyCoord(search_input[0], search_input[1], unit="deg")
    elif isinstance(search_input, str):
        # Try to turn the passed string into a SkyCoord.
        try:
            search_object = SkyCoord(search_input)
        except ValueError:
            # If this fails assume we were given an object name
            search_object = search_input
            # If no catalog is specified, see if we can infer a preffered catalog
            if catalog is None:
                search_catalog, _ = _match_target_catalog(search_input)
    else:
        raise TypeError("Cannot Resolve type of search input")

    # If a catalog is passed make this the catalog to search
    # Else use a default catalog(tic?) if none is assigned
    if isinstance(catalog, str):
        search_catalog = catalog
    elif catalog is None:
        search_catalog = None
    return search_object, search_catalog


def _match_target_catalog(search_input):
    search = search_input.strip().replace(" ", "").lower()
    if search_input.isnumeric():
        # If string is purelt numbers, make no assumptions
        search_string = search_input
        search_catalog = None
    elif search[0:3] == "tic":
        search_catalog = "tic"
        search_string = search[3:]
    elif search[0:4] == "tess":
        search_catalog = "tic"
        search_string = search[4:]
    elif search[0:3] == "kic":
        search_catalog = "kic"
        search_string = search[3:]
    elif search[0:4] == "kplr":
        search_catalog = "kic"
        search_string = search[4:]
    elif search[0:4] == "epic":
        search_catalog = "epic"
        search_string = search[4:]
    elif search[0:4] == "ktwo":
        search_catalog = "epic"
        search_string = search[4:]
    elif search[0:7] == "gaiadr3":
        search_catalog = "gaiadr3"
        search_string = search[7:]
    elif search[0:4] == "gaia" and search_input[4:6] != "dr":
        search_string = search[4:]
        search_catalog = "gaiadr3"
    else:
        # If we cannot parse a catalog, make no assumptions
        search_catalog = (None,)
        search_string = search_input
    return search_catalog, search_string


def _parse_id_list(search_object):
    ids = [_parse_id(item)[0] for item in search_object]
    id_list = ", ".join(str(id) for id in ids)
    return id_list


def _parse_id(search_item):
    if isinstance(search_item, int):
        id = str(search_item)
        scat = None
    elif isinstance(search_item, str):
        scat, id = _match_target_catalog(search_item)
    else:
        id = None
        scat = None

    return id, scat


def _get_return_columns(columns):
    """Convenience function to reorder columns and remove motion columns."""
    downselect_columns = list(
        set(columns)
        - set(
            [
                "ID",
                "RA",
                "Dec",
                "RAJ2000",
                "DEJ2000",
                "Plx",
                "pmRA",
                "pmDE",
                "Separation",
                "Relative_Flux",
            ]
        )
    )
    downselect_columns = np.hstack(
        [
            np.sort([i for i in downselect_columns if i.endswith("mag")]),
            np.sort([i for i in downselect_columns if not i.endswith("mag")]),
        ]
    )
    new_columns = [
        "ID",
        "RA",
        "Dec",
        "Separation",
        "Relative_Flux",
        "pmRA",
        "pmDE",
        *downselect_columns,
    ]
    return new_columns


def _table_to_skycoord(
    table: Table, equinox: Time = None, output_epoch: Time = None, catalog=None
) -> SkyCoord:
    """
    Convert a table input to astropy.coordinates.SkyCoord object

    Parameters
    ----------
    table : astropy.table.Table
        astropy.table.Table which contains the coordinates of targets and proper motion values
    equinox: astropy.time.Time
        The equinox for the catalog
    output_epoch : astropy.time.Time
        Desired time of the observation

    Returns
    -------
    coords : astropy.coordinates.SkyCoord
       SkyCoord object with RA, Dec, equinox, and proper motion parameters.
    """

    if equinox is None and catalog is None:
        _, catalog = _parse_id(table[0][0])
    if equinox is None and catalog is not None:
        equinox = _Catalog_Dictionary[catalog]["equinox"]
    if output_epoch is None and catalog is not None:
        output_epoch = equinox

    # We need to remove any nan values from our proper  motion list
    # Doing this will allow objects which do not have proper motion to still be displayed
    if "pmRA" in table.keys():
        table["pmRA"] = np.ma.filled(table["pmRA"].astype(float), 0.0)
    else:
        table["pmRA"] = 0.0 * u.mas / u.yr

    if "pmDE" in table.keys():
        table["pmDE"] = np.ma.filled(table["pmDE"].astype(float), 0.0)
    else:
        table["pmDE"] = 0.0 * u.mas / u.yr

    # If an object does not have a parallax then we treat it as if the object is an "infinite distance"
    # and set the parallax to 1e-7 arcseconds or 10Mpc.
    if "Plx" in table.keys():
        table["Plx"] = np.ma.filled(table["Plx"].astype(float), 1e-4)
    else:
        table["Plx"] = 1e-4 * u.mas

    if "RAJ2000" in table.keys() and "DEJ2000" in table.keys():
        ra = table["RAJ2000"]
        dec = table["DEJ2000"]
    elif "RA" in table.keys() and "Dec" in table.keys():
        ra = table["RA"] * u.deg
        dec = table["Dec"] * u.deg
    else:
        raise KeyError("RA, DEC Disambiguation failed")
    # Suppress warning caused by Astropy as noted in issue 111747 (https://github.com/astropy/astropy/issues/11747)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="negative parallaxes")

        # Get the input data from the table
        c = SkyCoord(
            ra=ra,
            dec=dec,
            distance=Distance(parallax=table["Plx"].quantity, allow_negative=True),
            pm_ra_cosdec=table["pmRA"],
            pm_dec=table["pmDE"],
            frame="icrs",
            obstime=equinox,
        )

    # Suppress warning caused by Astropy as noted in issue 111747 (https://github.com/astropy/astropy/issues/11747)
    if output_epoch != equinox:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="ERFA function")
            warnings.filterwarnings("ignore", message="invalid value")
            c = c.apply_space_motion(new_obstime=output_epoch)
    return c
